@@ -205,7 +205,7 @@ struct arguments {
 	struct atom_code {
 		size_t n;
 		int * routine;
-	} atom_code[MAX_ATOMS];
+	} *atom_code[MAX_ATOMS];
 	/* File names seen on the command line.  */
 	char const * fname[MAX_FNAMES];
 	/**/
@@ -233,21 +233,14 @@ struct atom * clone_atom(struct arguments * args)
 	rv->ignore_case = atom->ignore_case;
 	rv->pat = atom->pat;
 	rv->patlen = atom->patlen;
-	struct atom_code * oac = &args->atom_code[oa];
-	struct atom_code * nac = &args->atom_code[na];
-	assert(nac->n == 0);
-	assert(oac->n > 0);
-	nac->n = oac->n+2;
-	nac->routine = malloc(nac->n * sizeof *nac->routine);
-	if (nac->routine == 0) fatal_enomem(0);
-	for (size_t i = 0; i < oac->n; i++) {
-		nac->routine[i] = oac->routine[i];
-	}
-	nac->routine[oac->n+0] = I_PUSH(na);
-	nac->routine[oac->n+1] = I_OR;
-	struct atom_code tmp = *oac;
-	args->atom_code[oa] = *nac;
-	args->atom_code[na] = tmp;
+	struct atom_code * ac = args->atom_code[oa];
+	args->atom_code[na] = ac;
+	assert(ac->n > 0);
+	ac->n += 2;
+	ac->routine = realloc(ac->routine, ac->n * sizeof *ac->routine);
+	if (ac->routine == 0) fatal_enomem(0);
+	ac->routine[ac->n-2] = I_PUSH(na);
+	ac->routine[ac->n-1] = I_OR;
 	return rv;
 }
 
@@ -340,13 +333,15 @@ static struct atom * enter_atom(struct arguments * args)
 		fail();
 	}
 	APPTOK(args->p.num_atoms + TOK_ATOM_BASE);
-	assert(args->atom_code[args->p.num_atoms].n == 0);
-	args->atom_code[args->p.num_atoms].n = 1;
-	args->atom_code[args->p.num_atoms].routine = malloc(1 * sizeof(int));
-	if (args->atom_code[args->p.num_atoms].routine == 0) {
+	args->atom_code[args->p.num_atoms] =
+		malloc(sizeof *args->atom_code[args->p.num_atoms]);
+	if (args->atom_code[args->p.num_atoms] == 0) fatal_enomem(0);
+	args->atom_code[args->p.num_atoms]->n = 1;
+	args->atom_code[args->p.num_atoms]->routine = malloc(1 * sizeof(int));
+	if (args->atom_code[args->p.num_atoms]->routine == 0) {
 		fatal_enomem(0);
 	}
-	args->atom_code[args->p.num_atoms].routine[0] 
+	args->atom_code[args->p.num_atoms]->routine[0] 
 		= I_PUSH(args->p.num_atoms);
 	rv = &args->p.atoms[args->p.num_atoms++];
 	rv->field_name = 0;
@@ -640,7 +635,7 @@ static void parse_prim(struct arguments * args)
 	int atom = get_token(args) - TOK_ATOM_BASE;
 	assert(atom >= 0);
 	assert(atom < MAX_ATOMS);
-	struct atom_code *ac = &args->atom_code[atom];
+	struct atom_code *ac = args->atom_code[atom];
 	for (size_t i = 0; i < ac->n; i++) {
 		addinsn(&args->p, ac->routine[i]);
 	}
